import abc
from typing import TypeVar, Generic

T = TypeVar('T')


class YamlNavigator(abc.ABC, Generic[T]):
    @abc.abstractproperty
    def _raw(self) -> T:
        """
        Gets access to the underlying collection.
        :return:
        """
        raise Exception("Not implemented")

    @abc.abstractmethod
    def __len__(self) -> int:
        raise Exception("Not implemented")

    @abc.abstractmethod
    def __iter__(self):
        raise Exception("Not implemented")



from typing import List, Optional, Any
import copy
import yaml


# FIXME: move this to its own library: YamlDict seems a good name
from yamldict.YamlNavigator import YamlNavigator


class YamlList(YamlNavigator[List[Any]]):
    """
    A property navigator that allows accessing a list and
    correctly wraps potentially nested dictionaries.
    """
    def __init__(self,
                 *args,
                 content: Optional[List] = None,
                 property_name: str = ""):
        if args:
            raise Exception("You need to pass the argument names")

        super(YamlList, self).__init__()

        self.__content = content if content is not None else list()
        self.__property_name = property_name

    def __deepcopy__(self, memodict={}):
        return YamlList(property_name=self.__property_name,
                        content=copy.deepcopy(self.__content))

    def __getitem__(self, item):
        result = self.__content[item]

        if isinstance(result, dict):
            return YamlDict(
                property_name=f"{self.__property_name}.{item}",
                content=result)

        if isinstance(result, list):
            return YamlList(
                property_name=f"{self.__property_name}.{item}",
                content=result)

        return result

    def __setitem__(self, key, value):
        if isinstance(value, YamlNavigator):
            value = value._raw

        self.__content[key] = value

    def __delitem__(self, key):
        self.__content.__delitem__(key)

    def __iter__(self):
        return YamlIteratorWrapper(
            iter=self.__content.__iter__(),
            property_name=self.__property_name,
        )

    def __len__(self) -> int:
        return len(self.__content)

    @property
    def _raw(self) -> List:
        """
        Get access to the underlying collection.
        :return:
        """
        return self.__content

    def __repr__(self):
        return f"YamlList({self.__property_name}) {self.__content}"


from yamldict.YamlDict import YamlDict
from yamldict.YamlIteratorWrapper import YamlIteratorWrapper


def yaml_representer(dumper, data):
    return dumper.represent_data(data._raw)


yaml.add_representer(YamlList, yaml_representer, Dumper=yaml.SafeDumper)



# FIXME: move this to its own library: YamlDict seems a good name
from typing import Dict, Optional, Any
import copy
import yaml

from yaml import SafeDumper

from yamldict.YamlNavigator import YamlNavigator


class YamlDict(YamlNavigator[Dict[str, Any]]):
    """
    A property navigator that allows accessing a dictionary via
    properties.
    """
    def __init__(self,
                 *args,
                 content: Optional[Dict] = None,
                 property_name: Optional[str] = ""):
        if args:
            raise Exception("You need to pass the named arguments")

        super(YamlDict, self).__init__()

        self.__content = content if content is not None else dict()
        self.__property_name = property_name

    def __deepcopy__(self, memodict={}):
        return YamlDict(
            property_name=self.__property_name,
            content=copy.deepcopy(self._raw))

    def __getattr__(self, item):
        if item == '_YamlDict__content':
            return self.__content

        if item == '_YamlDict__property_name':
            return self.__property_name

        # FIXME I'm not sure why the YamlMissing gets called here even from the parent
        if item == '_YamlDict__create_if_missing' or item == '_YamlMissing__create_if_missing':
            return self.__create_if_missing

        if item not in self.__content:
            return YamlMissing(
                parent_property=self,
                property_name=item,
                full_property_name=f"{self.__property_name}.{item}")

        result = self.__content[item]

        if isinstance(result, dict):
            return YamlDict(
                property_name=f"{self.__property_name}.{item}",
                content=result)
        elif isinstance(result, list):
            return YamlList(
                property_name=f"{self.__property_name}.{item}",
                content=result)

        return result

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setattr__(self, key, value):
        if isinstance(value, YamlNavigator):
            value = value._raw

        if "_YamlDict__content" == key:
            super(YamlDict, self).__setattr__(key, value)
            return

        if "_YamlDict__property_name" == key:
            super(YamlDict, self).__setattr__(key, value)
            return

        self.__content[key] = value

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __delattr__(self, item):
        self.__content.__delitem__(item)

    def __delitem__(self, key):
        self.__content.__delitem__(key)

    def __create_if_missing(self) -> 'YamlDict':
        return self

    def __iter__(self):
        return YamlIteratorWrapper(
            iter=self.__content.__iter__(),
            property_name=self.__property_name,
        )

    def __len__(self) -> int:
        return len(self.__content)

    def _items(self):
        return YamlDictWrapper(
            items_view=self.__content.items(),
            property_name=self.__property_name,
        )

    @property
    def _raw(self) -> Dict:
        """
        Gets access to the underlying collection.
        :return:
        """
        return self.__content

    def __repr__(self) -> str:
        return f"YamlDict({self.__property_name}) {self.__content}"


from yamldict.YamlList import YamlList
from yamldict.YamlIteratorWrapper import YamlIteratorWrapper, YamlDictWrapper
from yamldict.YamlMissing import YamlMissing


def yaml_representer(dumper, data):
    return dumper.represent_data(data._raw)


yaml.add_representer(YamlDict, yaml_representer, Dumper=SafeDumper)



from typing import Iterator, Any, ItemsView


class YamlIteratorWrapper:
    def __init__(self,
                 *args,
                 property_name: str,
                 iter: Iterator):
        if args:
            raise Exception("You need to pass the arguments by name")

        self._property_name = property_name
        self._iter = iter

    def __iter__(self):
        return YamlIteratorWrapper(
            property_name=self._property_name,
            iter=self._iter
        )

    def __next__(self):
        result = self._iter.__next__()

        return convert_type(
            property_name=f"{self._property_name}._iterator",
            value=result)


class YamlDictIteratorWrapper:
    def __init__(self,
                 *args,
                 property_name: str,
                 iter: Iterator):
        if args:
            raise Exception("You need to pass the arguments by name")

        self._property_name = property_name
        self._iter = iter

    def __iter__(self):
        return YamlDictIteratorWrapper(
            property_name=self._property_name,
            iter=self._iter
        )

    def __next__(self):
        result = self._iter.__next__()

        return (
            result[0],
            convert_type(
                property_name=f"{self._property_name}.{result[0]}",
                value=result[1],
            ),
        )


class YamlDictWrapper:
    def __init__(self,
                 *args,
                 property_name: str,
                 items_view: ItemsView):
        if args:
            raise Exception("You need to pass the arguments by name")

        self._property_name = property_name
        self._items_view = items_view

    def __iter__(self):
        return YamlDictIteratorWrapper(
            property_name=self._property_name,
            iter=self._items_view.__iter__(),
        )


def convert_type(*args,
                 property_name: str,
                 value: Any) -> Any:
    if args:
        raise Exception("You need to pass the arguments by name")

    if isinstance(value, dict):
        return YamlDict(
            property_name=property_name,
            content=value)
    elif isinstance(value, list):
        return YamlList(
            property_name=property_name,
            content=value)

    return value


from yamldict.YamlDict import YamlDict
from yamldict.YamlList import YamlList



from setuptools import setup
from setuptools import find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

packages = find_packages()

setup(
    name='yamldict',
    version='1.0.0',
    description='yamldict',
    long_description=readme,
    author='Bogdan Mustiata',
    author_email='bogdan.mustiata@gmail.com',
    license='BSD',
    install_requires=[
        "PyYAML==5.1.2"],
    packages=packages,
    package_data={
        '': ['*.txt', '*.rst']
    })



from typing import Any, Union, List, Dict

from yamldict.YamlNavigator import YamlNavigator


class YamlMissing(YamlNavigator[None]):
    EMPTY_LIST: List[str] = list()

    def __init__(self,
                 *args,
                 parent_property: Union['YamlDict', 'YamlMissing'],
                 property_name: str,
                 full_property_name: str):
        if args:
            raise Exception("You need to pass named arguments")

        if property_name.endswith("__parent_property"):
            raise Exception("Problem")

        super(YamlMissing, self).__init__()

        self.__property_name = property_name
        self.__full_property_name = full_property_name
        self.__parent_property = parent_property

    def __getattr__(self, item):
        if item == '_YamlMissing__parent_property':
            return self.__parent_property

        if item == '_YamlMissing__property_name':
            return self.__property_name

        if item == '_YamlMissing__full_property_name':
            return self.__property_name

        if item == '_YamlMissing__create_if_missing':
            return self.__create_if_missing

        # If we get calls for other attributes, we just return none
        return YamlMissing(
            parent_property=self,
            property_name=item,
            full_property_name=f"{self.__full_property_name}.{item}")

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setattr__(self, key, value) -> None:
        if key == '_YamlMissing__property_name':
            super(YamlMissing, self).__setattr__(key, value)
            return

        if key == '_YamlMissing__parent_property':
            super(YamlMissing, self).__setattr__(key, value)
            return

        if key == '_YamlMissing__full_property_name':
            super(YamlMissing, self).__setattr__(key, value)
            return

        if self.__parent_property is None:
            raise Exception(f"No parent_property set on the missing property {self}")

        parent: YamlDict = self.__parent_property.__create_if_missing()

        if self.__property_name in parent:
            container = parent[self.__property_name]
        else:
            container = dict()
            parent[self.__property_name] = container

        container[key] = value

    def __create_if_missing(self) -> 'YamlDict':
        parent: YamlDict = self.__parent_property.__create_if_missing()

        if self.__property_name in parent:
            return parent[self.__property_name]

        container: Dict[str, Any] = dict()
        parent[self.__property_name] = container

        return YamlDict(
            property_name=self.__property_name,
            content=container,
        )

    def __setitem__(self, key, value) -> None:
        self.__setattr__(key, value)

    def _raw(self) -> Any:
        return None

    def __len__(self):
        return 0

    def __iter__(self):
        return YamlMissing.EMPTY_LIST.__iter__()

    def __repr__(self):
        return f"YamlMissing({self.__full_property_name})"

    def __deepcopy__(self, memodict={}):
        return YamlMissing(
            parent_property=None,  # FIXME: this means assignment won't work anymore
            property_name=self.__property_name,
            full_property_name=self.__full_property_name)


from yamldict.YamlDict import YamlDict
